package ru.dbotthepony.mc.otm.graph.storage

import net.minecraft.world.level.block.entity.BlockEntity
import ru.dbotthepony.mc.otm.capability.MatteryCapability
import ru.dbotthepony.mc.otm.capability.energy.IMatteryEnergyStorage
import ru.dbotthepony.mc.otm.graph.GraphNode
import ru.dbotthepony.mc.otm.storage.IStorage
import ru.dbotthepony.mc.otm.storage.IVirtualStorageComponent
import ru.dbotthepony.mc.otm.storage.StorageStack
import ru.dbotthepony.mc.otm.storage.VirtualComponent
import ru.dbotthepony.mc.otm.storage.powered.PoweredVirtualComponent
import java.util.*
import java.util.function.Supplier

open class StorageNode(private val energyDemander: IMatteryEnergyStorage? = null) : GraphNode<StorageNode, StorageGraph>(::StorageGraph) {
	protected val components = ArrayList<IStorage<*>>()

	/**
	 * Setting this to true will render non functional default [attachComponents],
	 * make [addStorageComponent] and [removeStorageComponent] not add/remove components
	 * to attached graph, and [revive] not re-attach components.
	 *
	 * [invalidate] and [removeComponents] still detach all components
	 */
	protected var manualAttaching = false
	private var demandingEnergy = false

	/**
	 * Called by storage graph on node being attached to it
	 */
	open fun attachComponents(to: StorageGraph) {
		if (energyDemander != null) {
			if (energyDemander.missingPower.isPositive) {
				demandingEnergy = true
				to.powerDemandingNodes.add(energyDemander)
			}
		}

		if (manualAttaching) {
			return
		}

		for (component in components) {
			to.add(component)
		}
	}

	/**
	 * Called by storage graph on node being detached from it
	 *
	 * This is NOT called when graph is being destroyed (e.g. even by merging
	 * with another graph).
	 */
	open fun removeComponents(from: StorageGraph) {
		for (component in components) {
			from.remove(component)
		}

		if (energyDemander != null) {
			from.powerDemandingNodes.remove(energyDemander)
			demandingEnergy = false
		}
	}

	fun tickEnergyDemanding() {
		energyDemander ?: throw IllegalStateException("No energy demander")

		if (!demandingEnergy && energyDemander.missingPower.isPositive) {
			demandingEnergy = true
			graph.powerDemandingNodes.add(energyDemander)
		} else if (demandingEnergy && !energyDemander.missingPower.isPositive) {
			demandingEnergy = false
			graph.powerDemandingNodes.remove(energyDemander)
		}
	}

	@Suppress("unchecked_cast")
	fun <T : StorageStack<T>, U : IStorage<T>> computeIfAbsent(type: StorageStack.Type<T>, provider: (StorageStack.Type<T>) -> U): U {
		return components.firstOrNull { it.storageType == type } as U? ?: provider(type).also { addStorageComponent(it) }
	}

	fun addStorageComponent(component: IStorage<*>) {
		if (!components.any { component === it || it.storageType === component.storageType }) {
			components.add(component)

			if (isValid && !manualAttaching)
				graph.add(component)
		}
	}

	fun removeStorageComponent(component: IStorage<*>) {
		val indexOf = components.indexOfFirst { component === it || it.storageType === component.storageType }
		if (indexOf == -1) return
		val self = components.removeAt(indexOf)
		if (isValid) graph.remove(self)
	}

	fun removeStorageComponent(component: StorageStack.Type<*>) {
		val indexOf = components.indexOfFirst { it.storageType === component }
		if (indexOf == -1) return
		val self = components.removeAt(indexOf)
		if (isValid) graph.remove(self)
	}

	override fun invalidate() {
		for (component in components) {
			graph.remove(component)
		}
	}

	override fun revive() {
		if (!manualAttaching) {
			for (component in components) {
				graph.add(component)
			}
		}
	}

	fun discover(blockEntity: BlockEntity) {
		discover(blockEntity, MatteryCapability.STORAGE_NODE)
	}
}
